/*
    Copyright 2006-2016 Luigi Auriemma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

    http://www.gnu.org/licenses/gpl-2.0.txt
*/

#ifdef DISABLE_COMPRESSION
#else
#define ENABLE_COMPRESSION  // comment this line for removing LZO compression!
#define MINILZO             // no boring dependencies
#endif

#define _LARGE_FILES        // if it's not supported the tool will work
#define MYSIZE  uint64_t    // correctly without problems but naturally
#define __USE_LARGEFILE64   // without support for large files
#define __USE_FILE_OFFSET64
#define _LARGEFILE_SOURCE
#define _LARGEFILE64_SOURCE
#define _FILE_OFFSET_BITS   64



#ifdef NOLFS                // disable 64 bits number
    #undef _LARGE_FILES
    #undef __USE_LARGEFILE64
    #undef __USE_FILE_OFFSET64
    #undef _LARGEFILE_SOURCE
    #undef _LARGEFILE64_SOURCE
    #undef _FILE_OFFSET_BITS
    #undef MYSIZE
    #define MYSIZE  uint32_t
#endif



#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <stdarg.h>
#include <sys/stat.h>
#include <ctype.h>
#include <time.h>
#include <fcntl.h>

#ifdef ENABLE_COMPRESSION
    #ifdef MINILZO
        #include "minilzo.h"    // minilzo is good too since it's the same
    #else
        #include <lzo/lzo1x.h>  // http://www.oberhumer.com/opensource/lzo/
    #endif
#endif
#include "md5.h"
#include "arc4.h"

#ifdef WIN32
    #include <winsock.h>
    #include <direct.h>
    #include "winerr.h"

    #define close       closesocket
    #define sleep       Sleep
    #define PATHSLASH   '\\'
    #define in_addr_t   uint32_t

    void winerr(DWORD error);
#else
    #include <unistd.h>
    #include <sys/socket.h>
    #include <sys/types.h>
    #include <arpa/inet.h>
    #include <netinet/in.h>
    #include <netdb.h>
    #include <dirent.h>
    #include <pthread.h>
    #include <utime.h>

    #define PATHSLASH   '/'
    #define stristr strcasestr
    #define stricmp strcasecmp
#endif

#ifdef WIN32
    #define quick_thread(NAME, ARG) DWORD WINAPI NAME(ARG)
    #define thread_id   HANDLE
#else
    #define quick_thread(NAME, ARG) void *NAME(ARG)
    #define thread_id   pthread_t
#endif

thread_id quick_threadx(void *func, void *data) {
    thread_id   tid;
#ifdef WIN32
    DWORD   tmp;

    tid = CreateThread(NULL, 0, func, data, 0, &tmp);
    if(!tid) return(0);
#else
    pthread_attr_t  attr;
    pthread_attr_init(&attr);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
    if(pthread_create(&tid, NULL, func, data)) return(0);
#endif
    return(tid);
}

typedef uint8_t     u8;
typedef uint16_t    u16;
typedef uint32_t    u32;
typedef uint64_t    u64;



#define VER             "0.1.6"
#define BUFFSZ          0xffff
#define COMPSZ          0xf0b1          // output = input + (input / 16) + 64 + 3
                                        // input  = ((output - 67) * 16) / 17
#define PASSWORDSZ      32
#define EXECIN          "#FILE"
#define EXECINSZ        (sizeof(EXECIN) - 1)
#define EXECOUTSZ       FNAMEMAX
#define YES             'y'
#define NO              'n'

#define FCLOSE(X)       if(fd && (fd != stdin) && (fd != stdout)) { fclose(fd); fd = NULL; }

u64 FILETIME2TIME(u64 X) {
    return (X / 10000000ULL) - 11644473600LL;
}
u64 TIME2FILETIME(u64 X) {
    return (X + 11644473600LL) * 10000000ULL;
}

#ifndef MAX_PATH
    #ifndef PATH_MAX
        #define MAX_PATH 2000
    #else
        #define MAX_PATH PATH_MAX
    #endif
#endif

#define FNAMEMAX        (MAX_PATH + FILENAME_MAX)

#define BAR_INIT        bar_step = filesize / 100;                          \
                        if(bar_step) {      /* normal (files > 100) */      \
                            bar_goal = 0;                                   \
                            bar_perc = 0;                                   \
                        } else {            /* trick (micro files) */       \
                            bar_step = 1;                                   \
                            bar_goal = filesize;                            \
                            bar_perc = 100;                                 \
                        }
#define BAR_SHOW        if(!quiet && (currsize >= bar_goal)) {              \
                            do {                                            \
                                bar_goal += bar_step;                       \
                                bar_perc++;                                 \
                            } while(bar_goal <= currsize);                  \
                            if(bar_perc > 100) bar_perc = 100;              \
                            printf("\r  %3d%%", bar_perc);                  \
                        }



#ifndef IP_TOS
    #define IP_TOS 3
#endif



#if defined(_LARGE_FILES)
    #if defined(__APPLE__)
        #define fseek   fseeko
        #define ftell   ftello
    #elif defined(__FreeBSD__)
    #elif !defined(NOLFS)       // use -DNOLFS if this tool can't be compiled on your OS!
        #define off_t   off64_t
        #define fopen   fopen64
        #define fseek   fseeko64
        #define ftell   ftello64
    #endif
#endif



enum {
    INIT_OK         = 1,
    INIT_PASSWORD   = 2,
    INIT_FASTMODE   = 4,
    INIT_SSL        = 8,

    INIT_ERROR      = 128
};

// 8 bit -> 32 bit
enum {
    FILE_OK         = 1,

    FILE_SKIP       = 2,
    FILE_RESUME     = 4,
    FILE_HASH       = 8,

    FILE_COMPRESS   = 16,
    FILE_CRYPT      = 32,
    FILE_XOR        = 64,

    FILE_ERROR      = 128,

    FILE_TIMESTAMP  = 256
};



// this is only a funny feature implemented on the fly,
// if it's interesting I will add a better real support in future
struct sockaddr_in  peer_udp;
#define UDP_BUFFSZ      1400

static const int
        on              = 1,
        off             = 0,
        tos             = 0x10;
int     overwrite       = 0,
        onlyone         = 0,
        forceskip       = 0,
        forceresume     = 0,
        fastmode        = 0,
        hashcheck       = 0,
        quiet           = 0,
        compress        = 0,
        xcrypt          = 0,
        simplexor       = 0,
        timestamp       = 0,
        nolimit         = 0,
        startarg        = 0,
        cmdargc         = 0,
        execlen         = 0,
        skip_fail       = 0,
        udp_mode        = 0;
u8      cdir[FNAMEMAX]  = "",
        **cmdargv       = NULL,
        *tmpexec        = NULL,
        *execstring     = NULL,
        *password       = NULL;



#include "ssl.h"
int client(int sd);
quick_thread(server, int sd);
int xsendfile(SSL *ssl_sd, int sd, u8 **buff, u8 **lzbuff, u8 **lzmem);
int xrecvfile(SSL *ssl_sd, int sd, u8 **buff, u8 **lzbuff);
int rootfix(u8 *dir);
int dir_checks(u8 *in, u8 *dirname, u8 **fname, u8 *dname);
int recursive_dir(SSL *ssl_sd, int sd, u8 *filedir, u8 **buff, u8 **lzbuff, u8 **lzmem);
int sendonefileonly(SSL *ssl_sd, int sd, u8 *fname, u8 **buff, u8 **lzbuff, u8 **lzmem);
int send_it(SSL *ssl_sd, int sd, u8 *fname, MYSIZE filesize, u8 **buff, u8 **lzbuff, u8 **lzmem);
int recv_mem(SSL *ssl_sd, int sd, u8 *data, int len);
MYSIZE recv_num(SSL *ssl_sd, int sd, int bits, int *err);
#define send_mem(SS,S,M,L)  mysend(SS, S, M, L)
int send_num(SSL *ssl_sd, int sd, u64 num, int bits);
in_addr_t *iplist(u8 *data);
void create_dir(u8 *name);
void show_examples(void);
u8 *get_str(u8 *data, int size);
int delimit(u8 *data);
int get_chr_answer(int num, ...);
void exec_arg(u8 *data);
void exec_run(u8 *fname);
void md5_password(u32 timex, u8 *pwd, u8 *md5h);
void md5_str(u8 *str, u8 *md5h);
MYSIZE mymd5_file(FILE *fd, u8 *md5h, MYSIZE reqsize);
void xorit(u8 *data, int len);
in_addr_t resolv(char *host);
void std_err(void);



int check_next_arg(int i, int argc, char **argv, int is_num) {
    u8      *p,
            c;

    i++;
    if(i >= argc) return(-1);
    p = argv[i];
    if(*p == '/') return(-1);
    if(*p == '-') {
        if(!is_num) return(-1);
        p++;  // for negative numbers
    }
    if(*p == '+') {
        if(is_num) return(0);
    }
    if(!is_num) return(0);

    c = tolower(*p);
    if((c >= '0') && (c <= '9')) return(0); // 0-9 covers also hex 0x

    return(-1);
}



int main(int argc, char *argv[]) {
    struct  linger  ling = {1,1};
    struct  sockaddr_in peer;
    in_addr_t   bindip,
                *allowip = NULL;
    int     sd,
            sda,
            i,
            psz,
            listenmode  = 0,
            interactive = 0;
    u16     port        = 6969;
    u8      *host       = NULL,
            *myfolder   = NULL,
            *p;

#ifdef WIN32
    WSADATA    wsadata;
    WSAStartup(MAKEWORD(1,0), &wsadata);
#endif

    setbuf(stdin,  NULL);
    setbuf(stdout, NULL);
    setbuf(stderr, NULL);

    fflush(stdin);  // useless?
    #ifdef O_BINARY
    setmode(fileno(stdin), O_BINARY);
    #endif

    fputs("\n"
        "Lanfile " VER "\n"
        "by Luigi Auriemma\n"
        "e-mail: aluigi@autistici.org\n"
        "web:    aluigi.org\n"
        "\n", stdout);

    getcwd(cdir, FNAMEMAX);

    if(argc < 2) {
        printf("\n"
            "Usage: %s [options] <host> [file/dir1] ... [file/dirN]\n"
            "\n"
            "Options:\n"
            "-p PORT   server or local port to use (%hu)\n"
            "-b IP     IP of the interface to bind (any interface)\n"
            "-d DIR    incoming folder (current: %s)\n"
            "-o        automatic files overwriting\n"
            "-k        automatic skipping all the existing files\n"
            "-r        automatic file resuming\n"
            "-m        MD5 hash verification only, no files transfer\n"
            "-f        FAST mode useful with many files, it automatically activates -o !\n"
            "-1        one full incoming transfer/client only and then exit\n"
            "-a I1,In  list of IPs and hostnames which are allowed to connect to you when\n"
            "          you are in listening mode, use a comma for separating the hosts\n"
            "-x \"EXE\"  launch a command for each file received. EXE is the command-line,\n"
            "          use the #FILE pattern which will be substituited with each file\n"
#ifdef ENABLE_COMPRESSION
            "-z        enable compression\n"
#endif
            "-c        rc4 encryption using the filename as key\n"
            "-C        simple XOR obfuscation\n"
            "-l        no limit size, use this option with devices and special files\n"
            "-q        quiet output, no file percentage\n"
            "-Q        show only the errors\n"
            "-i        interactive mode, you must simply answer to some questions\n"
            "-e        some practical examples of how to use this tool\n"
            "-X        don't quit in case of non critical errors, useful if a file can't be\n"
            "          read/written and to not stop the sending of many files just for one\n"
            "-t        send and set the original timestamps\n"
            "-U        experimental UDP mode, only one client supported\n"
            "\n"
            "Server-only options:\n"
            "-w PASS   set a password valid only in listening mode, in connect mode this\n"
            "          password will be used when connected to a passworded server\n"
#ifdef ENABLE_SSL
            "-S C P    enable SSL with certificate C and password P (no cert verify)\n"
#endif
            "\n"
            "Use the <host> 0 for the listening mode (accept incoming connections).\n"
            "If you specify files or folders they will be sent to the other host,\n"
            "specify no files for enabling the recvfile mode for receiving files, examples:\n"
            "  listen  & recv:  lanfile 0\n"
            "  listen  & send:  lanfile 0 c:\\myfolder c:\\myfile1.txt\n"
            "  connect & recv:  lanfile 192.168.0.1\n"
            "  connect & send:  lanfile 192.168.0.1 c:\\myfolder c:\\myfile1.txt\n"
            "You can also use wildcards (*file*) if your OS supports them or - for stdin.\n"
            "Empty folders are not transmitted.\n"
            "\n", argv[0], port, cdir);
        exit(1);
    }

    bindip = htonl(INADDR_ANY);

    for(i = 1; i < argc; i++) {
        if((argv[i][0] != '-') && (argv[i][0] != '/')) break;
        switch(argv[i][1]) {
            case 'p': port          = atoi(argv[++i]);      break;
            case 'b': bindip        = inet_addr(argv[++i]); break;
            case 'd': myfolder      = argv[++i];            break;
            case 'o': overwrite     = 1;                    break;
            case 'k': forceskip     = 1;                    break;
            case 'r': forceresume   = 1;                    break;
            case 'f': fastmode      = 1;                    break;
            case 'm': hashcheck     = 1;                    break;
            case '1': onlyone       = 1;                    break;
            case 'a': allowip       = iplist(argv[++i]);    break;
            case 'x': exec_arg(argv[++i]);                  break;
            case 'w': password      = argv[++i];            break;
#ifdef ENABLE_COMPRESSION
            case 'z': compress      = 1;                    break;
#endif
            case 'c': xcrypt        = 1;                    break;
            case 'C': simplexor     = 1;                    break;
            case 'l': nolimit       = 1;                    break;
            case 'q': quiet         = 1;                    break;
            case 'Q': quiet         = -1;                   break;
            case 'i': interactive   = 1;                    break;
            case 'e': show_examples();                      break;
            case 'X': skip_fail     = 1;                    break;
            case 't': timestamp     = 1;                    break;
            case 'S': {
                dossl           = 1;
                ssl_cert_file   = argv[++i];
                if(!ssl_cert_file[0]) ssl_cert_file = NULL;
                ssl_cert_pass   = argv[++i];
                if(!ssl_cert_pass[0]) ssl_cert_pass = NULL;
                break;
            }
            case 'U': udp_mode      = 1; onlyone = 1;       break;
            // don't use check_next_arg because there is no fixed number of mandatory arguments
            default: {
                printf("\nError: wrong command-line argument (%s)\n\n", argv[i]);
                exit(1);
                break;
            }
        }
    }

    host     = argv[i];
    if(i >= argc) host = NULL;
    if(host) {
        p = strchr(host, ':');
        if(p) {
            *p = 0;
            port = atoi(p + 1);
        }
    }
    startarg = ++i;
    cmdargc  = argc;
    cmdargv  = (void *)argv;    // avoids the warning

    if(interactive) {
        startarg = 0;
        cmdargv = malloc(sizeof(char *));
        if(!cmdargv) std_err();
        cmdargv[0] = malloc(FNAMEMAX);
        if(!cmdargv[0]) std_err();

        printf("- do you want to send or receive files? (s/r)\n  ");
        if(get_chr_answer(2, 's', 'r') == 's') {
            cmdargc = 1;
            printf("- is the destination able to receive incoming connections? (y/n)\n  ");
            if(get_chr_answer(2, YES, NO) == YES) {
                if(!host) {
                    printf("- insert the IP or the hostname of the destination:\n  ");
                    host = get_str(NULL, 0);
                }
            } else {
                printf("- the destination must connect to you for receiving the data\n");
                host = "0";
            }
            printf("- insert the name of the file or folder you want to send:\n  ");
            get_str(cmdargv[0], FNAMEMAX);
        } else {
            cmdargc = 0;
            printf("- can you receive incoming connections? (y/n)\n  ");
            if(get_chr_answer(2, YES, NO) == NO) {
                if(!host) {
                    printf("- insert the IP or the hostname of the source:\n  ");
                    host = get_str(NULL, 0);
                }
            } else {
                printf("- the source must connect to you for sending the data\n");
                host = "0";
            }
            getcwd(cdir, FNAMEMAX);
            printf(
                "- insert the path of the folder in which you want to receive\n"
                "  use . or hit RETURN for the current one (%s):\n  ",
                cdir);
            myfolder = get_str(NULL, 0);
            if(!myfolder[0]) strcpy(myfolder, ".");
        }

    } else {
        if(startarg > argc) {
            printf("\nError: you have not specified the host, use 0 for listening mode\n\n");
            exit(1);
        }
    }

    if(myfolder) {
        if(chdir(myfolder) < 0) std_err();
    }
    getcwd(cdir, FNAMEMAX);

    if(password) {
        if(strlen(password) > PASSWORDSZ) {
            password[PASSWORDSZ] = 0;
            printf("- password truncated at %d bytes\n", PASSWORDSZ);
        }
    }
    if(fastmode) {
        overwrite   = 1;
        forceskip   = 0;
        forceresume = 0;
        hashcheck   = 0;
        printf("- fast mode enabled\n");
    }

    if(overwrite)   printf("- force file overwriting\n");
    if(forceskip)   printf("- force file skipping\n");
    if(forceresume) printf("- force file resuming\n");
    if(hashcheck)   printf("- MD5 hash checking\n");
    if(onlyone)     printf("- one incoming connection only\n");
    if(compress)    printf("- compression enabled\n");
    if(xcrypt)      printf("- encryption enabled\n");
    if(simplexor)   printf("- XOR obfuscation enabled\n");
    if(timestamp)   printf("- timestamps restoring enabled\n");
    if(dossl)       printf("- SSL enabled\n");

    peer.sin_addr.s_addr = resolv(host);
    peer.sin_port        = htons(port);
    peer.sin_family      = AF_INET;

    if((peer.sin_addr.s_addr == INADDR_ANY) || (peer.sin_addr.s_addr == INADDR_NONE)) listenmode = 1;

    if(startarg < cmdargc) {
        printf("- SENDFILE mode:\n");
    } else {
        printf("- incoming folder:   %s\n", cdir);
        printf("- RECVFILE mode:\n");
    }

    if(listenmode) {
        printf("- LISTEN mode\n");
    } else {
        printf("- CONNECT mode\n");
    }

    if(udp_mode) sd = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
    else         sd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
    if(sd < 0) std_err();
    //setsockopt(sd, SOL_SOCKET, TCP_NODELAY, (char *)&on, sizeof(on));
    setsockopt(sd, SOL_SOCKET, SO_LINGER, (char *)&ling, sizeof(ling));
    setsockopt(sd, SOL_SOCKET, SO_BROADCAST, (char *)&on, sizeof(on));
    setsockopt(sd, IPPROTO_IP, IP_TOS, (char *)&tos, sizeof(tos));

    if(listenmode) {
        peer.sin_addr.s_addr = bindip;

        printf("- listen             %s : %hu\n", inet_ntoa(peer.sin_addr), ntohs(peer.sin_port));

        if(setsockopt(sd, SOL_SOCKET, SO_REUSEADDR, (char *)&on, sizeof(on))
          < 0) std_err();
        if(bind(sd, (struct sockaddr *)&peer, sizeof(struct sockaddr_in))
          < 0) std_err();
        if(!udp_mode) {
            if(listen(sd, SOMAXCONN)
              < 0) std_err();
        }

        printf("- clients:\n");

        for(;;) {
            psz = sizeof(struct sockaddr_in);
            if(udp_mode) {
                if(recvfrom(sd, (void *)&i, sizeof(i), 0, (struct sockaddr *)&peer, &psz) < 0) continue;
                memcpy(&peer_udp, &peer, sizeof(struct sockaddr_in));
                sda = sd;
            } else {
                sda = accept(sd, (struct sockaddr *)&peer, &psz);
                if(sd < 0) std_err();
            }
            //setsockopt(sda, SOL_SOCKET, TCP_NODELAY, (char *)&on, sizeof(on));
            setsockopt(sda, SOL_SOCKET, SO_LINGER, (char *)&ling, sizeof(ling));

            printf("  %s:%hu\n", inet_ntoa(peer.sin_addr), ntohs(peer.sin_port));

            if(allowip) {
                for(i = 0; allowip[i]; i++) {
                    if(peer.sin_addr.s_addr == allowip[i]) break;
                }
                if(!allowip[i]) {
                    printf("  not allowed\n");
                    close(sda);
                    continue;
                }
            }

            if(!quick_threadx(server, (void *)sda)) close(sda);

            if(onlyone) {
                for(;;) sleep(-1);
            }
        }

    } else {

        printf("- target             %s : %hu\n", inet_ntoa(peer.sin_addr), ntohs(peer.sin_port));

        printf("- connect...");
        if(udp_mode) {
            if(sendto(sd, "", 0, 0, (struct sockaddr *)&peer, sizeof(struct sockaddr_in))
              < 0) std_err();
            memcpy(&peer_udp, &peer, sizeof(struct sockaddr_in));
        } else {
            if(connect(sd, (struct sockaddr *)&peer, sizeof(struct sockaddr_in))
              < 0) std_err();
        }
        printf("ok\n");

        if(client(sd) < 0) goto quit;
    }

quit:
    close(sd);
    if(interactive) {
        free(cmdargv[0]);
        free(cmdargv);
        printf("\n  Press RETURN to quit\n");
        fflush(stdin);
        u8  ans[16];
        fgets(ans, sizeof(ans), stdin);
    }
    return(0);
}



FILE *myfopen(u8 *fname, u8 *mode) {
    if(!fname || !mode) return NULL;
    if(!strcmp(fname, "-")) {
        if(strchr(mode, 'r')) {
            return stdin;
        } else {
            return stdout;
        }
    }
    return fopen(fname, mode);
}



/*
void bar_init(MYSIZE filesize) {
    bar_step = filesize / 100;
        if(bar_step) {      // normal (files > 100)
        bar_goal = 0;
        bar_perc = 0;
    } else {            // trick (micro files)
        bar_step = 1;
        bar_goal = filesize;
        bar_perc = 100;
    }
}



void bar_show(MYSIZE currsize) {
    if(!quiet && (currsize >= bar_goal)) {
        do {
            bar_goal += bar_step;
            bar_perc++;
        } while(bar_goal <= currsize);
        if(bar_perc > 100) bar_perc = 100;
        printf("\r  %3d%%", bar_perc);
    }
}
*/



void showit(MYSIZE num, u8 *fname, int skip) {
    int     i;
    u8      nums[22];

    nums[14] = 0;
    for(i = 13; num && (i >= 0); i--) {
        nums[i] = '0' + (num % 10);
        num /= 10;
    }
    for(; i >= 0; i--) {
        nums[i] = ' ';
    }
    if(quiet >= 0) printf("       %s %s%s", nums, skip ? "SKIP " : "", fname);
}



int client(int sd) {
#ifdef ENABLE_SSL
    SSL_CTX     *ctx_sd     = NULL;
#endif
    SSL         *ssl_sd     = NULL;

    u32     timex;
    time_t  start_time;
    int     err,
            ret         = -1;
    u32     init8,
            res_opt;
    u8      password_md5[16],
            *buff       = NULL,
            *lzbuff     = NULL,
            *lzmem      = NULL;

    init8 = recv_num(ssl_sd, sd, 8, &err);                  // receive init byte
    if(err < 0) goto quit;
    if(init8 == 0xff) {
        init8 = recv_num(ssl_sd, sd, 32, &err);
        if(err < 0) goto quit;
    }
    if(!(init8 & INIT_OK)) {
        printf("\n  Error: your request has not been accepted");
        goto quit;
    }
    if(init8 & INIT_SSL) {
#ifdef ENABLE_SSL
        ssl_init();
        ctx_sd = SSL_CTX_new(SSL_METHOD);
        if(!ctx_sd) goto quit;
        //SSL_COMPAT(ctx_sd)
        ssl_sd = SSL_new(ctx_sd);
        SSL_set_fd(ssl_sd, sd);
        if(SSL_connect(ssl_sd) < 0) goto quit;
#else
        printf("\n  Error: SSL not supported in this build");
        goto quit;
#endif
    }
    if(init8 & INIT_PASSWORD) {
        timex = recv_num(ssl_sd, sd, 32, &err);             // current server time
        if(err < 0) goto quit;
        if(!password) {
            printf("- server is protected with password, insert the keyword:\n  ");
            password = get_str(NULL, PASSWORDSZ + 1);
            md5_password(timex, password, password_md5);
            free(password);
        } else {
            md5_password(timex, password, password_md5);
        }
        if(send_mem(ssl_sd, sd, password_md5, 16) < 0) {    // my password hash
            goto quit;
        }
        res_opt = recv_num(ssl_sd, sd, 8, &err);
        if(err < 0) goto quit;
        if(res_opt == 0xff) {
            res_opt = recv_num(ssl_sd, sd, 32, &err);
            if(err < 0) goto quit;
        }
        if(!(res_opt & INIT_OK)) {                  // if wrong password
            printf("\n  Error: wrong password");
            goto quit;
        }
    }
    if(init8 & INIT_FASTMODE) {
        if(!fastmode) {
            printf("\n  Error: the server uses the fast mode, you must enable it with -f");
            goto quit;
        }
    }

    start_time = time(NULL);
    if(startarg < cmdargc) {
        xsendfile(ssl_sd, sd, &buff, &lzbuff, &lzmem);
    } else {
        xrecvfile(ssl_sd, sd, &buff, &lzbuff);
    }
    printf("\n- transfer time:     %u seconds\n", (u32)(time(NULL) - start_time));
    ret = 0;

quit:
    if(buff) free(buff);
    if(lzmem) free(lzmem);
    if(lzbuff) free(lzbuff);
    return(ret);
}



quick_thread(server, int sd) {
#ifdef ENABLE_SSL
    SSL_CTX     *ctx_sd     = NULL;
#endif
    SSL         *ssl_sd     = NULL;

    time_t  start_time;
    u32     timex;
    u32     init8;
    u8      pass[16],
            password_md5[16],
            *buff       = NULL,
            *lzbuff     = NULL,
            *lzmem      = NULL;

    init8 = INIT_OK;
    if(password) init8 |= INIT_PASSWORD;
    if(fastmode) init8 |= INIT_FASTMODE;
    if(dossl)    init8 |= INIT_SSL;
    if(init8 > 0xff) {
        if(send_num(ssl_sd, sd, 0xff, 8)    < 0) goto quit;     // send init
        if(send_num(ssl_sd, sd, init8, 32)  < 0) goto quit;     // send init
    } else {
        if(send_num(ssl_sd, sd, init8, 8)   < 0) goto quit;     // send init
    }

    if(init8 & INIT_SSL) {
#ifdef ENABLE_SSL
        ssl_init();
        ctx_sd = SSL_CTX_new(SSL_METHOD);
        if(!ctx_sd) goto quit;
        //SSL_COMPAT(ctx_sd)

        if(ssl_cert_pass) SSL_CTX_set_default_passwd_cb(ctx_sd, pem_passwd_cb);
        if(ssl_cert_file) {
            int j = -1;
            if(SSL_CTX_use_certificate_chain_file(ctx_sd, ssl_cert_file) == 1) {
                j = 0;
                if(SSL_CTX_use_PrivateKey_file(ctx_sd, ssl_cert_file, SSL_FILETYPE_PEM)   != 1) \
                if(SSL_CTX_use_PrivateKey_file(ctx_sd, ssl_cert_file, SSL_FILETYPE_ASN1)  != 1) \
                if(SSL_CTX_use_certificate_file(ctx_sd, ssl_cert_file, SSL_FILETYPE_PEM)  != 1) \
                if(SSL_CTX_use_certificate_file(ctx_sd, ssl_cert_file, SSL_FILETYPE_ASN1) != 1) j = -1;
            }
            if(j < 0) {
                fprintf(stderr, "\n"
                    "Error: problems with the loading of the certificate file\n"
                    "       check if the certificate you specified is in PEM format and the\n"
                    "       password for the private key and the choosed SSL method are correct\n"
                    "- the following is a quick example for creating a quick certificate:\n"
                    "   openssl req -x509 -days 365 -newkey rsa:1024 -keyout cert.pem -out cert.pem\n"
                    "  add -nodes for a passwordless certificate\n");
                exit(1);
            }
        } else {
            if(!SSL_CTX_use_certificate_ASN1(ctx_sd, sizeof(SSL_CERT_X509) - 1, SSL_CERT_X509) ||
               !SSL_CTX_use_PrivateKey_ASN1(EVP_PKEY_RSA, ctx_sd, SSL_CERT_RSA, sizeof(SSL_CERT_RSA) - 1)) {
                fprintf(stderr, "\nError: problems with the loading of the certificate in memory\n");
                exit(1);
            }
        }
        SSL_CTX_set_verify_depth(ctx_sd, 1);  // #if (OPENSSL_VERSION_NUMBER < 0x00905100L)

        ssl_sd = SSL_new(ctx_sd);
        if(!ssl_sd) goto quit;
        SSL_set_fd(ssl_sd, sd);
        if(SSL_accept(ssl_sd) < 0) goto quit;
#else
        printf("\n  Error: SSL not supported in this build");
        goto quit;
#endif
    }

    if(init8 & INIT_PASSWORD) {
        timex = time(NULL);
        if(send_num(ssl_sd, sd, timex, 32)  < 0) goto quit;     // send current time
        if(recv_mem(ssl_sd, sd, pass,  16)  < 0) goto quit;     // client's password hash
        md5_password(timex, password, password_md5);
        if(memcmp(pass, password_md5, 16)) {
            send_num(ssl_sd, sd, INIT_ERROR, 8);                // error = no INIT_OK
            goto quit;
        }
        if(send_num(ssl_sd, sd, INIT_OK, 8) < 0) goto quit;     // let's go = INIT_OK
    }

    start_time = time(NULL);
    if(startarg < cmdargc) {
        xsendfile(ssl_sd, sd, &buff, &lzbuff, &lzmem);
    } else {
        xrecvfile(ssl_sd, sd, &buff, &lzbuff);
    }
    printf("\n- transfer time:     %u seconds\n", (u32)(time(NULL) - start_time));

quit:
#ifdef ENABLE_SSL
    if(ssl_sd) {
        SSL_shutdown(ssl_sd);
        SSL_free(ssl_sd);
    }
    if(ctx_sd) {
        SSL_CTX_free(ctx_sd);
    }
#endif
    close(sd);
    if(buff) free(buff);
    if(lzmem) free(lzmem);
    if(lzbuff) free(lzbuff);
    if(onlyone) exit(1);
    return(0);
}



int xsendfile(SSL *ssl_sd, int sd, u8 **buff, u8 **lzbuff, u8 **lzmem) {
    int     i,
            err,
            ret     = -1;
    u8      sdir[FNAMEMAX],
            dname[FNAMEMAX],
            *fname;

    for(i = startarg; i < cmdargc; i++) {
        dir_checks(cmdargv[i], sdir, &fname, dname);
        chdir(sdir);
        if(fname) {
            err = sendonefileonly(ssl_sd, sd, fname, buff, lzbuff, lzmem);
        } else {
            err = recursive_dir(ssl_sd, sd, dname, buff, lzbuff, lzmem);
        }
        chdir(cdir);
        if(fname) free(fname);
        if(err < 0) goto quit;
    }
    ret = 0;
quit:
    return ret;
}



int xrecvfile(SSL *ssl_sd, int sd, u8 **buff, u8 **lzbuff) {
#ifdef WIN32
    HANDLE  fh;
#else
    struct utimbuf tvp;
#endif
    arc4_context rc4_ctx;
    md5_context md5t;
    FILE    *fd;
    MYSIZE  totfilesize,
            filesize,
            myfilesize,
            currsize,
            bar_step    = 0,
            bar_goal    = 0;
    int     i,
            len,
            lzlen,
            lzerr,
            err,
            namelen,
            skip,
            resume,
            mycompress,
            mycrypt,
            myxor,
            bar_perc    = 0,
            ret         = -1;
    u32     recv_opt,   // u8 + bigger size
            send_opt,
            res_opt;
    u8      fname[FNAMEMAX],
            origfname[FNAMEMAX],
            md5h[16],
            origmd5h[16];

    if(!*buff) {
        *buff = malloc(BUFFSZ);
        if(!*buff) std_err();
    }

    for(;;) {
        fd         = NULL;    // initialization
        skip       = 0;
        resume     = 0;
        currsize   = 0;

        namelen = recv_num(ssl_sd, sd, 16, &err);
        if(err < 0) goto quit;
        if(!namelen) break;
        if(namelen >= sizeof(fname)) {
            printf("\n- Error: the host has sent a too big filename");
            goto quit;
        }

        if(recv_mem(ssl_sd, sd, fname, namelen) < 0) {
            printf("\n- Error: filename has not been received");
            goto quit;
        }
        fname[namelen] = 0;
        strcpy(origfname, fname);

        totfilesize = filesize = recv_num(ssl_sd, sd, 64, &err);
        if(err < 0) goto quit;
        showit(filesize, fname, 0);

        if(
          (fname[0] == '/')     ||
          (fname[0] == '\\')    ||
          strstr(fname, "\\\\") ||
          strstr(fname, "//")   ||
          strstr(fname, "..\\") ||
          strstr(fname, "../")  ||
          strchr(fname, ':')) {
            printf("\n- Error: directory traversal");
            goto quit;
        }

        create_dir(fname);

        fd = myfopen(fname, "rb");
        if(fd) {
            if(forceskip) {
                skip = 1;
                fputc('\r', stdout);
                showit(filesize, fname, 1);
            } else if(overwrite) {
                // nothing to do
            } else if(forceresume) {
                resume = 1;
            } else if(hashcheck) {
                // nothing to do
            } else {
                printf("\n- do you want to Skip / Overwrite / Resume this file? (s/o/r)\n  ");
                switch(get_chr_answer(3+3, 's', 'o', 'r', /*general keys*/ 'n','y','a')) {
                    case 'n':
                    case 's': {
                        skip = 1;
                        showit(filesize, fname, 1);
                        break;
                    }
                    case 'y':
                    case 'o': {
                        showit(filesize, fname, 0);
                        break;
                    }
                    case 'a':
                    case 'r': {
                        resume = 1;
                        showit(filesize, fname, 0);
                        break;
                    }
                    default: break;
                }
            }
            FCLOSE(fd);
        } else {
            // if needed in future
        }

        recv_opt = FILE_OK;
        if(skip)      recv_opt |= FILE_SKIP;
        if(resume)    recv_opt |= FILE_RESUME;
        if(hashcheck) recv_opt |= FILE_HASH;
        if(compress)  recv_opt |= FILE_COMPRESS;
        if(xcrypt)    recv_opt |= FILE_CRYPT;
        if(simplexor) recv_opt |= FILE_XOR;
        if(timestamp) recv_opt |= FILE_TIMESTAMP;
        if(fastmode) {
            // nothing to do
        } else {
            if(recv_opt > 0xff) {
                if(send_num(ssl_sd, sd, 0xff, 8) < 0) goto quit;
                len = send_num(ssl_sd, sd, recv_opt, 32);
            } else {
                len = send_num(ssl_sd, sd, recv_opt, 8);
            }
            if(len < 0) {     // recvfile option
                printf("\n- Error: the host has closed the connection");
                goto quit;
            }
        }

        send_opt = recv_num(ssl_sd, sd, 8, &err);       // sendfile option
        if(err < 0) goto quit;
        if(send_opt == 0xff) {
            send_opt = recv_num(ssl_sd, sd, 32, &err);  // sendfile option
            if(err < 0) goto quit;
        }
        if(send_opt & FILE_ERROR) {
            printf("\n- Error: the host has not accepted your request");
            goto quit;
        }

        if(skip) {
            fputc('\n', stdout);
            continue;

        } else if(resume) {
            fd = myfopen(fname, "rb");
            if(!fd) {
                perror("\n- Error");
                if(skip_fail) continue;
                goto quit;
            }
            myfilesize = mymd5_file(fd, md5h, -1);            // -1 means all
            FCLOSE(fd);

            if(send_num(ssl_sd, sd, myfilesize, 64) < 0) goto quit; // send local file size
            if(send_mem(ssl_sd, sd, md5h, 16)       < 0) goto quit; // send our temporary MD5
            res_opt = recv_num(ssl_sd, sd, 8, &err);
            if(err < 0) goto quit;
            if(res_opt == 0xff) {
                res_opt = recv_num(ssl_sd, sd, 32, &err);
                if(err < 0) goto quit;
            }
            if(res_opt & FILE_RESUME) {                     // receive status, if resume or not
                fd = myfopen(fname, "ab");
                currsize = myfilesize;
            } else {
                if(quiet >= 0) printf("\rCANNOT RESUME *\n");
                fd = myfopen(fname, "wb");
            }

        } else if((recv_opt & FILE_HASH) || (send_opt & FILE_HASH)) {
            fd = myfopen(fname, "rb");
            if(!fd) {
                myfilesize = -1;
                memset(md5h, 0, sizeof(md5h));
            } else {
                myfilesize = mymd5_file(fd, md5h, -1);
                FCLOSE(fd);
            }
            if(send_num(ssl_sd, sd, myfilesize, 64) < 0) goto quit; // send local file size
            if(send_mem(ssl_sd, sd, md5h, 16)       < 0) goto quit; // send our temporary MD5
            res_opt = recv_num(ssl_sd, sd, 8, &err);
            if(err < 0) goto quit;
            if(res_opt == 0xff) {
                res_opt = recv_num(ssl_sd, sd, 32, &err);
                if(err < 0) goto quit;
            }
            if(res_opt & FILE_OK) {
                if(quiet >= 0) printf("\rOK\n");
            } else {
                if(quiet >= 0) printf("\r!!!WRONG!!!\n");
            }
            continue;

        } else {
            fd = myfopen(fname, "wb");
        }

        if(!fd) {
            perror("\n- Error");
            if(!skip_fail) goto quit;
        }

        md5_starts(&md5t);

        if(quiet) {
            fputc('\n', stdout);
        } else {
            BAR_INIT;
        }

        mycompress = ((recv_opt & FILE_COMPRESS) || (send_opt & FILE_COMPRESS));
        mycrypt    = ((recv_opt & FILE_CRYPT)    || (send_opt & FILE_CRYPT));
        myxor      = ((recv_opt & FILE_XOR)      || (send_opt & FILE_XOR));

        if(mycrypt) {
            arc4_setup(&rc4_ctx, origfname, namelen);
        }
        //setsockopt(sd, SOL_SOCKET, TCP_NODELAY, (char *)&off, sizeof(off));

#ifdef ENABLE_COMPRESSION
        if(mycompress) {
            if(!*lzbuff) {
                *lzbuff = malloc(BUFFSZ);
                if(!*lzbuff) std_err();
            }

            for(filesize -= currsize; filesize; filesize -= len) {
                lzlen = recv_num(ssl_sd, sd, 16, &err);
                if(err < 0) goto quit;
                if(recv_mem(ssl_sd, sd, *lzbuff, lzlen) < 0) {
                    if(totfilesize != -1) printf("\n- Error: the host has closed the connection");
                    goto quit;
                }

                if(myxor)                  xorit(*lzbuff, lzlen);
                if(mycrypt) arc4_crypt(&rc4_ctx, *lzbuff, lzlen);

                len = BUFFSZ;
                lzerr = lzo1x_decompress_safe(*lzbuff, lzlen, *buff, (void *)&len, LZO1X_MEM_DECOMPRESS);
                if(lzerr != LZO_E_OK) {
                    printf("\n"
                        "- Error: LZO decompression error (%d)\n"
                        "         probably the input data is corrupted", lzerr);
                    if(!skip_fail) goto quit;
                }

                if(fd) {
                    if(fwrite(*buff, len, 1, fd) != 1) {
                        printf("\n- Error: I/O error, the file is incomplete or the disk space is finished");
                        goto quit;
                    }
                }

                md5_update(&md5t, *buff, len);

                currsize += len;
                BAR_SHOW;
            }

        } else
#endif
        {
            for(filesize -= currsize; filesize; filesize -= len) {
                len = myrecv(ssl_sd, sd, *buff, (filesize < BUFFSZ) ? filesize : BUFFSZ);
                if(len <= 0) {
                    if(totfilesize != -1) printf("\n- Error: the host has closed the connection");
                    goto quit;
                }

                if(myxor)                  xorit(*buff, len);
                if(mycrypt) arc4_crypt(&rc4_ctx, *buff, len);

                if(fd) {
                    if(fwrite(*buff, len, 1, fd) != 1) {
                        printf("\n- Error: I/O error, the file is incomplete or the disk space is finished");
                        goto quit;
                    }
                }

                md5_update(&md5t, *buff, len);

                currsize += len;
                BAR_SHOW;
            }
        }

        //setsockopt(sd, SOL_SOCKET, TCP_NODELAY, (char *)&on, sizeof(on));

        if(!quiet) fputc('\n', stdout);

        md5_finish(&md5t, md5h);

        if(recv_mem(ssl_sd, sd, origmd5h, 16) < 0) {
            printf("\n- Error: MD5 hash has not been received");
            goto quit;
        }

        if(memcmp(origmd5h, md5h, 16)) {
            printf("\n"
                "- Error: remote file differs from yours:\n"
                "    Original: ");
            for(i = 0; i < 16; i++) printf("%02x", origmd5h[i]);
            printf("\n    Your:     ");
            for(i = 0; i < 16; i++) printf("%02x", md5h[i]);
            printf("\n");
            if(!skip_fail) goto quit;
        }

        FCLOSE(fd);

        if((recv_opt & FILE_TIMESTAMP) || (send_opt & FILE_TIMESTAMP)) {
            u64 creation_time   = recv_num(ssl_sd, sd, 64, &err);
            if(err < 0) goto quit;
            u64 lastaccess_time = recv_num(ssl_sd, sd, 64, &err);
            if(err < 0) goto quit;
            u64 lastwrite_time  = recv_num(ssl_sd, sd, 64, &err);
            if(err < 0) goto quit;
            if((creation_time <= 0xffffffffLL) && (lastaccess_time <= 0xffffffffLL) && (lastwrite_time <= 0xffffffffLL)) {  // time_t
#ifdef WIN32
                creation_time   = TIME2FILETIME(creation_time);
                lastaccess_time = TIME2FILETIME(lastaccess_time);
                lastwrite_time  = TIME2FILETIME(lastwrite_time);
#endif
            } else {    // FILETIME
#ifndef WIN32
                tvp.actime  = FILETIME2TIME(lastaccess_time);
                tvp.modtime = FILETIME2TIME(lastwrite_time);
#endif
            }
#ifdef WIN32
            fh = CreateFile(fname, FILE_WRITE_ATTRIBUTES, FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
            if(fh != INVALID_HANDLE_VALUE) {
                SetFileTime(fh, (FILETIME *)&creation_time, (FILETIME *)&lastaccess_time, (FILETIME *)&lastwrite_time);
                CloseHandle(fh);
            }
#else
            // creation time is not available
            tvp.actime  = FILETIME2TIME(lastaccess_time);
            tvp.modtime = FILETIME2TIME(lastwrite_time);
            utime(fname, &tvp);
#endif
        }

        if(execstring) exec_run(fname);
    }

    ret = 0;
quit:
    //setsockopt(sd, SOL_SOCKET, TCP_NODELAY, (char *)&on, sizeof(on));
    FCLOSE(fd);
    return ret;
}



int rootfix(u8 *dir) {
    int     len;
    u8      *p;

    len = strlen(dir);

    p = strchr(dir, ':');
    if(!p) return(len);

    p++;
    if((*p != '\\') && (*p != '/')) {
        len++;
        memcpy(p + 1, p, len - (p - dir));
        *p = PATHSLASH;
    }
    return(len);
}



int dir_checks(u8 *in, u8 *dirname, u8 **fname, u8 *dname) {
    int     len;
    u8      *folder,
            *p;

    dname[0] = 0;
    *fname = NULL;

    strcpy(dirname, in);
    len = rootfix(dirname);

    for(p = dirname + len - 1; p >= dirname; p--) {
        if((*p != '\\') && (*p != '/')) break;
    }
    *(p + 1) = 0;

    folder = strrchr(dirname, '\\');
    if(!folder) folder = strrchr(dirname, '/');
    
    if(chdir(dirname) < 0) {
        if(folder) {
            *folder++ = 0;
            *fname = strdup(folder);
            rootfix(dirname);
            if(!dirname[0]) strcpy(dirname, "/");
            if(chdir(dirname) < 0) std_err();
        } else {
            *fname = strdup(dirname);
        }
    } else {
        chdir(cdir);
        if(folder) {
            *folder++ = 0;
            if(folder[0]) {
                strcpy(dname, folder);
            } else {
                strcpy(dname, ".");
            }
            rootfix(dirname);
            if(!dirname[0]) sprintf(dirname, "%c", PATHSLASH);
            if(chdir(dirname) < 0) std_err();
        } else {
            strcpy(dname, dirname);
        }
    }

    getcwd(dirname, FNAMEMAX);

    len = strlen(dirname) - 1;
    if((dirname[len] != '\\') && (dirname[len] != '/')) {
        dirname[len + 1] = PATHSLASH;
        dirname[len + 2] = 0;
    }

    return 0;
}



    /* THE TOOL DOESN'T SEND EMPTY FOLDERS! */
int recursive_dir(SSL *ssl_sd, int sd, u8 *filedir, u8 **buff, u8 **lzbuff, u8 **lzmem) {
    int     plen,
            ret     = -1;

#ifdef WIN32
    static int      winnt = -1;
    OSVERSIONINFO   osver;
    WIN32_FIND_DATA wfd;
    HANDLE  hFind;
    MYSIZE  fsize;

    if(winnt < 0) {
        osver.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
        GetVersionEx(&osver);
        if(osver.dwPlatformId >= VER_PLATFORM_WIN32_NT) {
            winnt = 1;
        } else {
            winnt = 0;
        }
    }

    plen = strlen(filedir);
    strcpy(filedir + plen, "\\*.*");
    plen++;

    if(winnt) { // required to avoid problems with Vista and Windows7!
        hFind = FindFirstFileEx(filedir, FindExInfoStandard, &wfd, FindExSearchNameMatch, NULL, 0);
    } else {
        hFind = FindFirstFile(filedir, &wfd);
    }
    if(hFind == INVALID_HANDLE_VALUE) return(0);
    do {
        if(!strcmp(wfd.cFileName, ".") || !strcmp(wfd.cFileName, "..")) continue;

        strcpy(filedir + plen, wfd.cFileName);

        if(wfd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
            if(recursive_dir(ssl_sd, sd, filedir, buff, lzbuff, lzmem) < 0) goto quit;
        } else {
            fsize = wfd.nFileSizeHigh;
            fsize = (fsize << 32) | wfd.nFileSizeLow;
            if(send_it(ssl_sd, sd, filedir, fsize, buff, lzbuff, lzmem) < 0) goto quit;
        }
    } while(FindNextFile(hFind, &wfd));
    ret = 0;

quit:
    FindClose(hFind);
#else
    struct  stat    xstat;
    struct  dirent  **namelist;
    int     n,
            i;

    n = scandir(filedir, &namelist, NULL, NULL);
    if(n < 0) {
        if(stat(filedir, &xstat) < 0) {
            printf("**** %s", filedir);
            if(skip_fail) return(0);
            std_err();
        }
        if(send_it(ssl_sd, sd, filedir, xstat.st_size, buff, lzbuff, lzmem) < 0) return(-1);
        return(0);
    }

    plen = strlen(filedir);
    strcpy(filedir + plen, "/");
    plen++;

    for(i = 0; i < n; i++) {
        if(!strcmp(namelist[i]->d_name, ".") || !strcmp(namelist[i]->d_name, "..")) continue;
        strcpy(filedir + plen, namelist[i]->d_name);

        if(stat(filedir, &xstat) < 0) {
            printf("**** %s", filedir);
            if(skip_fail) return(0);
            std_err();
        }
        if(S_ISDIR(xstat.st_mode)) {
            if(recursive_dir(ssl_sd, sd, filedir, buff, lzbuff, lzmem) < 0) goto quit;
        } else {
            if(send_it(ssl_sd, sd, filedir, xstat.st_size, buff, lzbuff, lzmem) < 0) goto quit;
        }
        free(namelist[i]);
    }
    ret = 0;

quit:
    for(; i < n; i++) free(namelist[i]);
    free(namelist);
#endif
    filedir[plen - 1] = 0;
    return(ret);
}



int sendonefileonly(SSL *ssl_sd, int sd, u8 *fname, u8 **buff, u8 **lzbuff, u8 **lzmem) {
#ifdef WIN32
    HANDLE  hFile;
    DWORD   high,
            low,
            error;
    MYSIZE  fsize;

    if(!strcmp(fname, "-")) return(send_it(ssl_sd, sd, fname, -1, buff, lzbuff, lzmem)); // stdin

    hFile = CreateFile(fname, 
        FILE_READ_ATTRIBUTES,
        FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE, 
        NULL,
        OPEN_EXISTING, 
        FILE_ATTRIBUTE_NORMAL, 
        NULL);
    if(hFile == INVALID_HANDLE_VALUE) {
        error = GetLastError();
        goto quit;
    }

    low = GetFileSize(hFile, &high);
    if((low == 0xffffffff) && (error = GetLastError()))  {
        goto quit;
    }
    CloseHandle(hFile);

    fsize = high;
    fsize = (fsize << 32) | low;
    return(send_it(ssl_sd, sd, fname, fsize, buff, lzbuff, lzmem));

quit:
    printf("**** %s", fname);
    winerr(error);
    if(skip_fail) return(0);
    exit(1);
    return(-1);     // avoids warnings
#else
    struct  stat    xstat;

    if(!strcmp(fname, "-")) return(send_it(ssl_sd, sd, fname, -1, buff, lzbuff, lzmem)); // stdin

    if(stat(fname, &xstat) < 0) {
        printf("**** %s", fname);
        perror("\nError: ");
        if(skip_fail) return(0);
        exit(1);
    }
    return(send_it(ssl_sd, sd, fname, xstat.st_size, buff, lzbuff, lzmem));
#endif
}



int send_it(SSL *ssl_sd, int sd, u8 *fname, MYSIZE filesize, u8 **buff, u8 **lzbuff, u8 **lzmem) {
    arc4_context rc4_ctx;
    md5_context md5t;
    FILE    *fd;
    MYSIZE  currsize    = 0,
            myfilesize,
            recv_filesize,
            bar_step    = 0,
            bar_goal    = 0;
    int     len,
            lzlen,
            lzerr,
            err,
            mycompress,
            mycrypt,
            myxor,
            mynolimit   = nolimit,
            bar_perc    = 0,
            ret         = -1;
    u32     recv_opt,
            send_opt;
    u16     namelen;
    u8      md5h[16],
            recv_md5h[16];

    showit(filesize, fname, 0);

    if(!strcmp(fname, "-")) {
        fd = stdin;
        fname = "stdin";
        mynolimit = 1;
    } else {
        fd = myfopen(fname, "rb");
        if(!fd) {
            if(skip_fail) return(0);
            std_err();
        }
    }

    namelen = strlen(fname);
    if(send_num(ssl_sd, sd, namelen,  16)   < 0) goto quit;
    if(send_mem(ssl_sd, sd, fname, namelen) < 0) goto quit;

    if(mynolimit) filesize = -1;
    if(send_num(ssl_sd, sd, filesize, 64)   < 0) goto quit;

    if(fastmode) {
        recv_opt = FILE_OK;
    } else {
        recv_opt = recv_num(ssl_sd, sd, 8, &err);
        if(err < 0) goto quit;
        if(recv_opt == 0xff) {
            recv_opt = recv_num(ssl_sd, sd, 32, &err);
            if(err < 0) goto quit;
        }
    }

    send_opt = FILE_OK;
    if(hashcheck) send_opt |= FILE_HASH;
    if(compress)  send_opt |= FILE_COMPRESS;
    if(xcrypt)    send_opt |= FILE_CRYPT;
    if(simplexor) send_opt |= FILE_XOR;
    if(timestamp) send_opt |= FILE_TIMESTAMP;
    if(send_opt > 0xff) {
        if(send_num(ssl_sd, sd, 0xff, 8) < 0) goto quit;
        if(send_num(ssl_sd, sd, send_opt, 32) < 0) goto quit;
    } else {
        if(send_num(ssl_sd, sd, send_opt, 8) < 0) goto quit;
    }

    if(recv_opt & FILE_SKIP) {
        if(quiet >= 0) printf("\r* SKIPPING *\n");
        ret = 0;
        goto quit;
    }
    if(recv_opt & FILE_RESUME) {
        recv_filesize = recv_num(ssl_sd, sd, 64, &err);         // <= client's file size
        if(err < 0) goto quit;
        if(recv_mem(ssl_sd, sd, recv_md5h, 16) < 0) goto quit;  // <= client's MD5 of the piece of file
        myfilesize = mymd5_file(fd, md5h, recv_filesize);
        if((myfilesize != recv_filesize) || memcmp(md5h, recv_md5h, 16)) {
            if(quiet >= 0) printf("\rCANNOT RESUME *\n");
            if(fd != stdin) FCLOSE(fd);                 // 64 bit compatibility (rewind)?
            fd = myfopen(fname, "rb");
            if(!fd) std_err();
            if(send_num(ssl_sd, sd, FILE_ERROR, 8)  < 0) goto quit;
        } else {
            if(send_num(ssl_sd, sd, FILE_RESUME, 8) < 0) goto quit;
            currsize = myfilesize;                      // resume
        }
    }

    if((recv_opt & FILE_HASH) || (send_opt & FILE_HASH)) {
        recv_filesize = recv_num(ssl_sd, sd, 64, &err);         // <= client's file size
        if(err < 0) goto quit;
        if(recv_mem(ssl_sd, sd, recv_md5h, 16) < 0) goto quit;  // <= client's MD5 of the file
        if(recv_filesize == -1) {                       // a bit faster with unex. files
            myfilesize = 0;
            memset(md5h, 0, sizeof(md5h));
        } else {
            myfilesize = mymd5_file(fd, md5h, -1);
        }
        if((myfilesize != recv_filesize) || memcmp(md5h, recv_md5h, 16)) {
            if(quiet >= 0) printf("\r!!!WRONG!!!\n");
            if(send_num(ssl_sd, sd, FILE_ERROR, 8) < 0) goto quit;
        } else {
            if(quiet >= 0) printf("\rOK\n");
            if(send_num(ssl_sd, sd, FILE_OK, 8) < 0) goto quit;
        }
        ret = 0;
        goto quit;
    }


    md5_starts(&md5t);

    if(quiet) {
        fputc('\n', stdout);
    } else {
        BAR_INIT;
    }

    mycompress = ((recv_opt & FILE_COMPRESS) || (send_opt & FILE_COMPRESS));
    mycrypt    = ((recv_opt & FILE_CRYPT)    || (send_opt & FILE_CRYPT));
    myxor      = ((recv_opt & FILE_XOR)      || (send_opt & FILE_XOR));

    if(!*buff) {
        *buff = malloc(BUFFSZ);
        if(!*buff) std_err();
    }
    if(mycrypt) {
        arc4_setup(&rc4_ctx, fname, namelen);
    }
    //setsockopt(sd, SOL_SOCKET, TCP_NODELAY, (char *)&off, sizeof(off));

#ifdef ENABLE_COMPRESSION
    if(mycompress) {
        if(!*lzbuff) {
            *lzbuff = malloc(BUFFSZ);
            if(!*lzbuff) std_err();
        }
        if(!*lzmem) {
            *lzmem = malloc(LZO1X_MEM_COMPRESS);
            if(!*lzmem) std_err();
        }

        while((len = fread(*buff, 1, udp_mode ? UDP_BUFFSZ : COMPSZ, fd))) {
            md5_update(&md5t, *buff, len);
            currsize += len;

            lzerr = lzo1x_1_compress(*buff, len, *lzbuff, (void *)&lzlen, *lzmem);
            if(lzerr != LZO_E_OK) {
                printf("\n  Error: LZO compression error (%d)", lzerr);
                if(!skip_fail) goto quit;
            }

            if(mycrypt) arc4_crypt(&rc4_ctx, *lzbuff, lzlen);
            if(myxor)                  xorit(*lzbuff, lzlen);

            if(send_num(ssl_sd, sd, lzlen,   16)    < 0) goto quit;
            if(send_mem(ssl_sd, sd, *lzbuff, lzlen) < 0) goto quit;

            BAR_SHOW;
        }

    } else
#endif
    {
        while((len = fread(*buff, 1, udp_mode ? UDP_BUFFSZ : BUFFSZ, fd))) {
            md5_update(&md5t, *buff, len);
            currsize += len;

            if(mycrypt) arc4_crypt(&rc4_ctx, *buff, len);
            if(myxor)                  xorit(*buff, len);

            if(send_mem(ssl_sd, sd, *buff, len) < 0) goto quit;

            BAR_SHOW;
        }
    }
    if(!quiet) fputc('\n', stdout);

    if(currsize < filesize) {
        memset(*buff, 0, BUFFSZ);
        for(;;) {
            len = filesize - currsize;
            if(!len) break;
            if(len > BUFFSZ) len = BUFFSZ;
            if(send_mem(ssl_sd, sd, *buff, len) < 0) goto quit;
            currsize += len;
        }
    }

    md5_finish(&md5t, md5h);

    if(!mynolimit) {
        if(send_mem(ssl_sd, sd, md5h, 16) < 0) goto quit;
    }

    if((recv_opt & FILE_TIMESTAMP) || (send_opt & FILE_TIMESTAMP)) {
        u64     creation_time   = 0;
        u64     lastaccess_time = 0;
        u64     lastwrite_time  = 0;

#ifdef WIN32
        HANDLE  fh;
        fh = CreateFile(fname, FILE_READ_ATTRIBUTES, FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
        if(fh != INVALID_HANDLE_VALUE) {
            GetFileTime(fh, (FILETIME *)&creation_time, (FILETIME *)&lastaccess_time, (FILETIME *)&lastwrite_time);
            CloseHandle(fh);
        } else
#endif
        {
            struct stat xstat;
            if(!stat(fname, &xstat) || !fstat(fileno(fd), &xstat)) {
                creation_time   = xstat.st_mtime;   // yeah mtime because there is no creation time! (st_ctime stands for change time)
                lastwrite_time  = xstat.st_mtime;
                lastaccess_time = xstat.st_atime;
            }
        }
        if(send_num(ssl_sd, sd, creation_time,   64) < 0) goto quit;
        if(send_num(ssl_sd, sd, lastaccess_time, 64) < 0) goto quit;
        if(send_num(ssl_sd, sd, lastwrite_time,  64) < 0) goto quit;
    }

    ret = 0;
quit:
    //setsockopt(sd, SOL_SOCKET, TCP_NODELAY, (char *)&on, sizeof(on));
    if(ret < 0) {
        printf("\n- Error: the host has closed the connection");
    }
    FCLOSE(fd);
    return(ret);
}



int recv_mem(SSL *ssl_sd, int sd, u8 *data, int len) {
    int     t;

    while(len) {
        t = myrecv(ssl_sd, sd, data, len);
        if(t <= 0) return(-1);
        data += t;
        len  -= t;
    }
    return(0);
}



u64 recv_num(SSL *ssl_sd, int sd, int bits, int *err) {
    u64     num,
            t;      // required for 64 bit compatibility!
    int     i,
            bytes;
    u8      tmp[8];

    bytes = bits >> 3;
    num  = 0;
    *err = -1;

    if(recv_mem(ssl_sd, sd, tmp, bytes) < 0) {
        return(num);
    }

    for(i = 0; i < bytes; i++) {
        t = tmp[i];
        num |= (t << (u64)(i << 3));
    }

    *err = 0;
    return(num);
}



int send_num(SSL *ssl_sd, int sd, u64 num, int bits) {
    int     i,
            bytes;
    u8      tmp[8];

    bytes = bits >> 3;
    for(i = 0; i < bytes; i++) {
        tmp[i] = num & 0xff;
        num >>= 8;
    }

    if(mysend(ssl_sd, sd, tmp, bytes)
      < 0) return(-1);
    return(0);
}



in_addr_t *iplist(u8 *data) {
    struct  in_addr tmp;
    in_addr_t   *list;
    int     i;
    u8      *p;

    printf("- create database of allowed hosts:\n");

    i = 1;
    for(p = data; *p; p++) {
        if(*p == ',') i++;
    }

    list = malloc(sizeof(in_addr_t) * i);
    if(!list) std_err();

    i = 0;
    for(;;) {
        p = strchr(data, ',');
        if(p) *p = 0;
        printf("  %s ... ", data);
        list[i] = resolv(data);
        tmp.s_addr = list[i];
        printf("%s\n", inet_ntoa(tmp));
        i++;
        if(!p) break;
        data = p + 1;
    }
    list[i] = 0;

    return(list);
}



void create_dir(u8 *name) {
    u8      *p;

    for(p = name; *p; p++) {
        if((*p == '/') || (*p == '\\')) {
            *p = 0;
    #ifdef WIN32
            mkdir(name);
    #else
            mkdir(name, 0755);
    #endif
            *p = PATHSLASH;
        }
    }
}



void show_examples(void) {
    printf(
        "\n"
        "Quick examples for using this tool\n"
        "You want to send the file test.txt and the folder myfolder to your friend\n"
        "You have the IP 7.7.7.7 and he has the IP 1.2.3.4\n"
        "\n"
        "- you and your friend can accept connections, no firewalls or routers:\n"
        "  you:   lanfile 1.2.3.4 test.txt myfolder\n"
        "  he:    lanfile 0\n"
        "\n"
        "- you cannot accept connections:\n"
        "  follow the above example\n"
        "\n"
        "- your friend cannot accept connections but you can:\n"
        "  you:   lanfile 0 test.txt myfolder\n"
        "  he:    lanfile 7.7.7.7\n"
        "\n"
        "- you and your friend cannot accept connections:\n"
        "  no way...\n"
        "\n"
        "- you want to send all the files with the word xyz in their name:\n"
        "  (your OS must support the wildcard feature, Win9x doesn't)\n"
        "  you:   lanfile 1.2.3.4 *xyz*\n"
        "  he:    lanfile 0\n"
        "\n"
        "- example of -x option, unzip each file received:\n"
        "         lanfile -x \"unzip -v #FILE\" 0\n"
        "\n"
        "- compression and obfuscation can be enabled with the -z and -C options:\n"
        "         lanfile -z -C 0\n"
        "\n"
        "- interactive mode for using this tool with some answers:\n"
        "         lanfile -i\n"
        "\n");
    exit(1);
}



u8 *get_str(u8 *data, int size) {
    if(!data) {
        if(!size) size = 256;
        data = malloc(size);
        if(!data) std_err();
    }
    fflush(stdin);
    fgets(data, size, stdin);
    delimit(data);
    return(data);
}



int delimit(u8 *data) {
    u8      *p;

    for(p = data; *p && (*p != '\n') && (*p != '\r'); p++);
    *p = 0;
    return(p - data);
}



int get_chr_answer(int num, ...) {
    va_list ap;
    int     i,
            chr,
            *av,
            res = -1;
    u8      ans[16];

    av = malloc(num * sizeof(int));
    if(!av) std_err();
    va_start(ap, num);
    for(i = 0; i < num; i++) {
        av[i] = tolower(va_arg(ap, int));
    }
    va_end(ap);

    fflush(stdin);
    do {
        if(!fgets(ans, sizeof(ans), stdin)) break;
        chr = tolower(ans[0]);
        if((chr == '\n') || (chr == '\r')) {
            res = av[0];    // default answer
        } else {
            for(i = 0; i < num; i++) {
                if(chr == av[i]) res = chr;
            }
        }
    } while(res < 0);

    free(av);
    return(res);
}



void exec_arg(u8 *data) {
    int     i;
    u8      *p;

    execstring = data;
    for(p = execstring, i = 0; (p = strstr(p, EXECIN)); p++, i++);

    tmpexec = malloc(strlen(execstring) - (i * EXECINSZ) + (i * EXECOUTSZ) + 1);
    if(!tmpexec) std_err();
}



void exec_run(u8 *fname) {
    int     fnamelen;
    u8      *p,
            *l,
            *execptr;

    fnamelen = strlen(fname);
    execptr  = tmpexec;

    for(p = execstring; (l = strstr(p, EXECIN)); p = l + EXECINSZ) {
        memcpy(execptr, p, l - p);
        execptr += l - p;
        memcpy(execptr, fname, fnamelen);
        execptr += fnamelen;
    }
    strcpy(execptr, p);

    printf("  Execute: \"%s\"\n", tmpexec);
    system(tmpexec);
}



void md5_password(u32 timex, u8 *pwd, u8 *md5h) {
    md5_context md5t;

    md5_starts(&md5t);
    md5_update(&md5t, (u8 *)&timex, 4);
    md5_update(&md5t, pwd, strlen(pwd));
    md5_finish(&md5t, md5h);
}



void md5_str(u8 *str, u8 *md5h) {
    md5_context md5t;

    md5_starts(&md5t);
    md5_update(&md5t, str, strlen(str));
    md5_finish(&md5t, md5h);
}



MYSIZE mymd5_file(FILE *fd, u8 *md5h, MYSIZE reqsize) {
    md5_context md5t;
    MYSIZE  filesize;
    int     len;
    u8      buff[4096];

    if(!fd) {
        memset(md5h, 0, 16);
        return(0);
    }
    md5_starts(&md5t);
    for(filesize = 0; reqsize; filesize += len, reqsize -= len) {
        len = fread(buff, 1, (reqsize < sizeof(buff)) ? reqsize : sizeof(buff), fd);
        if(!len) break;
        md5_update(&md5t, buff, len);
    }
    md5_finish(&md5t, md5h);
    return(filesize);
}



void xorit(u8 *data, int len) {
    u8  *limit;

    for(limit = data + len; data < limit; data++) {
        *data ^= 0xff;
    }
}



in_addr_t resolv(char *host) {
    struct      hostent *hp;
    in_addr_t   host_ip;

    host_ip = inet_addr(host);
    if(host_ip == htonl(INADDR_NONE)) {
        hp = gethostbyname(host);
        if(!hp) {
            printf("\nError: Unable to resolve hostname (%s)\n", host);
            exit(1);
        } else host_ip = *(in_addr_t *)hp->h_addr;
    }
    return(host_ip);
}



#ifdef WIN32
    void winerr(DWORD error) {
        char    *message;

        if(FormatMessage(
          FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
          NULL,
          error,
          0,
          (char *)&message,
          0,
          NULL)
        ) {
            printf("\n  Error: %s", message);
            LocalFree(message);
        }
    }
#else
    void std_err(void) {
        perror("\nError");
        exit(1);
    }
#endif


